﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using System.IO;

namespace TGM3Sram
{
    public class TGM3Sram
    {
        MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();

        public void Start(string[] args)
        {
            if (args.Length != 3)
                Usage();
            else if (args[0] == "d")
                Decrypt(args[1], args[2]);
            else if (args[0] == "e")
                Encrypt(args[1], args[2]);
            else
                Usage();
        }

        private void Usage()
        {
            Console.WriteLine("usage:");
            Console.WriteLine("TGM3Sram [d|e] in out");
        }

        private void Decrypt(string fileIn, string fileOut)
        {
            FileStream fs = new FileStream(fileIn, FileMode.Open);
            BinaryReader br = new BinaryReader(fs);

            uint c0 = br.ReadUInt32();
            uint c1 = br.ReadUInt32();
            uint c2 = br.ReadUInt32();
            uint c3 = br.ReadUInt32();

            uint h0 = c0 ^ 0xFA769BC1;
            uint h1 = h0 ^ c3;
            uint h2 = h1 ^ c2;
            uint h3 = h2 ^ c1;

            uint k0 = h3 ^ h0;
            uint k1 = h3;
            uint k2 = h2;
            uint k3 = h1;

            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);
            bw.Write(k0);
            bw.Write(k1);
            bw.Write(k2);
            bw.Write(k3);

            FileStream fs2 = new FileStream(fileOut, FileMode.Create);
            BinaryWriter bw2 = new BinaryWriter(fs2);

            Crypt(br, bw2, fs.Length - fs.Position, ms.ToArray());

            bw2.Close();
            fs2.Close();

            br.Close();
            fs.Close();
        }

        private void Encrypt(string fileIn, string fileOut)
        {
            FileStream fs2 = new FileStream(fileOut, FileMode.Create);
            BinaryWriter bw2 = new BinaryWriter(fs2);

            var fs = File.ReadAllBytes(fileIn);
            var sum = md5.ComputeHash(fs);

            MemoryStream ms = new MemoryStream(fs);
            BinaryReader br = new BinaryReader(ms);

            MemoryStream ms2 = new MemoryStream(sum);
            BinaryReader br2 = new BinaryReader(ms2);

            uint h0 = br2.ReadUInt32();
            uint h1 = br2.ReadUInt32();
            uint h2 = br2.ReadUInt32();
            uint h3 = br2.ReadUInt32();

            uint a = h0 ^ h1;

            uint k0 =  a ^ 0xFA769BC1;
            uint k1 = h1 ^ h2;
            uint k2 = h2 ^ h3;
            uint k3 = h3 ^ a;

            bw2.Write(k0);
            bw2.Write(k1);
            bw2.Write(k2);
            bw2.Write(k3);

            Crypt(br, bw2, fs.Length, sum);

            bw2.Close();
            fs2.Close();
        }

        private void Crypt(BinaryReader br, BinaryWriter bw2, long len, byte[] key)
        {
            byte change = 2;
            int pos = 0;

            while (pos < len)
            {
                change++;

                if (change == 255)
                    change = 0;

                key[1] ^= change;

                key = md5.ComputeHash(key);

                int i=0;
                while (i < 16)
                {
                    if (pos < len)
                    {
                        bw2.Write((byte)(key[i] ^ br.ReadByte()));
                        i++;
                        pos++;
                    }
                    else
                        break;
                }
            }
        }
    }
}
